// RayMap.cpp: implementation of the CRayMap class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MapEdit03.h"
#include "RayMap.h"
#include "math.h"		//do obliczenia wartosci funkcji trygonometrycznych

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// Konstruktor klasy CRayMap
// parametry inicjowane jest domylnymi wartociami
CRayMap::CRayMap()
{

int i=0;					
int j=0;
float angle;

WORLD_COLUMNS=16;		//rozmiar wiata ( w komorkach)
WORLD_ROWS=16;
CELL_X_SIZE=64;				//rozmiary komorek
CELL_Y_SIZE=64;
//rad_to_degree=float(3.141592653589793/180);

resolution_x=320;
rozmiar_tablicy=resolution_x*360/60;  // tablica bedzie zapelniana w zaleznosci od rozdzielczosci
delta_kata=float(360)/rozmiar_tablicy;// kwant kata
world_x_size=CELL_X_SIZE*WORLD_COLUMNS; // rozmiar swiata
world_y_size=CELL_Y_SIZE*WORLD_ROWS;

ANGLE_30=int(float(30)/360*rozmiar_tablicy);
ANGLE_90=int(float(90)/360*rozmiar_tablicy);
ANGLE_180=int(float(180)/360*rozmiar_tablicy);
ANGLE_270=int(float(270)/360*rozmiar_tablicy);
ANGLE_360=rozmiar_tablicy;

for(i=0;i<800;i++)bound_array[i]=0;

//Podstawowa mapa to prostokt - Utworzenie sciany dookola pomieszczenia => tablica z 1 dookola
	for(i=0;i<WORLD_ROWS;i++)
	for(j=0;j<WORLD_COLUMNS;j++) world[i][j]=0;
	for(i=0;i<WORLD_ROWS;i++) {world[i][0]=1;world[i][WORLD_COLUMNS-1]=2;}
	for(j=0;j<WORLD_COLUMNS;j++) {world[0][j]=3;world[WORLD_ROWS-1][j]=4;}
	world[3][3]=2;
// Inicjalizacja tablic trygonometrycznych

for(i=0;i<rozmiar_tablicy;i++) 
	{
		
		angle=float(3.14159/180*i*delta_kata);

		if (sin(angle)==0) inv_sin_table[i]=1e+8;
		else inv_sin_table[i]=(float)(1/sin(angle));
		
		if (cos(angle)==0) inv_cos_table[i]=1e+8;
		else inv_cos_table[i]=(float)(1/cos(angle));
		
		tan_table[i]=float(tan(angle));

		fabs_inv_cos_table[i]=(float)fabs(inv_cos_table[i]);
		if (tan_table[i]==0) inv_tan_table[i]=1e+8;
		else inv_tan_table[i]=(float)(1/tan_table[i]);

	}

for(i=0;i<ANGLE_90;i++) 
	{
	x_step[i]=inv_tan_table[i]*CELL_X_SIZE;
	y_step[i]=tan_table[i]*CELL_Y_SIZE;
	}
for(i=ANGLE_90;i<ANGLE_180;i++) 
	{
	x_step[i]=inv_tan_table[i]*CELL_X_SIZE;
	y_step[i]=-tan_table[i]*CELL_Y_SIZE;
	}
for(i=ANGLE_180;i<ANGLE_270;i++) 
	{
	x_step[i]=-inv_tan_table[i]*CELL_X_SIZE;
	y_step[i]=-tan_table[i]*CELL_Y_SIZE;
	}
for(i=ANGLE_270;i<rozmiar_tablicy;i++) 
	{
	x_step[i]=-inv_tan_table[i]*CELL_X_SIZE;
	y_step[i]=tan_table[i]*CELL_Y_SIZE;
	}


// Inicjalizacja korekcji grafiki

double rad_angle;
for(i=-160;i<160;i++)
	{
	rad_angle = (3.272e-5) + i * 2*3.141592654/rozmiar_tablicy;
	correction_table[i+160] = (float)(1/cos(rad_angle));
	}
/*
 char temp[56]; 
FILE *plik=fopen("c:\wyniki.txt","wt");
 for(i=0;i<1920;i++)
 {
	 sprintf(temp,"%d inv_sin=%4.2f inv_cos=%4.2f inv_tan=%4.2f",i,inv_sin_table[i],inv_cos_table[i],inv_tan_table[i]);
	 fwrite(&temp,sizeof(char),strlen(temp),plik);
	 sprintf(temp,"\n");
	fwrite(&temp,sizeof(char),strlen(temp),plik);
 }
 fclose(plik);

*/

}

// Destruktor klasy CRayMap
  CRayMap::~CRayMap()
{

}




// Funkcja ray_casting realizuje algorytm rzucania promieni
void CRayMap::ray_casting(long x, long y, int _view_angle,long scale)
{
bool xray=false;				//do sprawdzania czy zaszlo przeciecie w poziomie
bool yray=false;				//i w pionie

int cell_x;					//komorka w ktorej znajduje sie gracz
long cell_y;
int y_bound;					//wspolrzedne kolejnego przeciecia poziomego i pionowego
long x_bound;
long xb_save;					//kopie powyzszych wspolrzednych
long yb_save;
long x_delta;					//odleglosc o jaka nalezy sie przesunac aby znalezc sie 
long y_delta;					//w nastepnej komorce
long next_x_cell;				//uzywane do znalezienia cwiartki w ktorej znajduje sie promien
long next_y_cell;


float xi;						//uzywane do sledzenia przeciec poziomych i pionowych
float yi;
int xi_save;					//kopie powyzszych wspolrzednych
int yi_save;
float dist_x;					//odleglosc od przeciecia pionowego
float dist_y;		
int ray;						//promien-zmienna w petli for

int bitmap_x=1;					//rodzaj tekstury dla przeciecia poziomego i pionowego
int bitmap_y=1;
float syi;
float sxi;
float step_x;
float step_y;

int view_angle;
view_angle=int(6*float(320)/360*_view_angle);
//view_angle=_view_angle;
view_angle-=ANGLE_30;
/* Czesc 1	Normalizacja kata */

if(view_angle <0) view_angle+=ANGLE_360;

for (ray=resolution_x-1;ray>=0;ray--)
{

xray=false;
yray=false;

/* Czesc 2 Obliczenie pierwszego poziomego przeciecia => wyznaczenie xi */

step_x=-inv_tan_table[view_angle]*CELL_X_SIZE;
step_y=-tan_table[view_angle]*CELL_X_SIZE;

if (view_angle==0)
{
	dist_x=oblicz_dla_0(x,y,ray);
	dist_y=1e+8;
	xray=yray=true;
	bound_array[ray]=CELL_Y_SIZE/2;
	
}
else if (view_angle==ANGLE_180)
{
	dist_x=oblicz_dla_180(x,y,ray);
	dist_y=1e+8;
	xray=yray=true;
	bound_array[ray]=CELL_Y_SIZE/2;
}
else if (view_angle==ANGLE_90)
{
	dist_y=oblicz_dla_90(x,y,ray);
	dist_x=1e+8;
	xray=yray=true;
	bound_array[ray]=CELL_X_SIZE/2;
}
else if (view_angle==ANGLE_270)
{
	dist_y=oblicz_dla_270(x,y,ray);
	dist_x=1e+8;
	xray=yray=true;
	bound_array[ray]=CELL_X_SIZE/2;}


if(view_angle>0 && view_angle<ANGLE_180)  // jesli gracz patrzy do gory
{
	y_bound=(y/CELL_Y_SIZE+1)*CELL_Y_SIZE;
	y_delta=CELL_Y_SIZE;
	xi=inv_tan_table[view_angle]*(y_bound-y)+x;
	next_y_cell=0;
	
}
else //patrzy w dol
{
	y_bound=(y/CELL_Y_SIZE)*CELL_Y_SIZE;
	y_delta=-CELL_Y_SIZE;
	xi=inv_tan_table[view_angle]*(y_bound-y)+x;
	next_y_cell=-1;
} // Koniec czesci 2

/* Czesc 3 Obliczenie pierwszego pionowego przeciecia => wyznaczenie xi */
if(view_angle<ANGLE_90 || view_angle>ANGLE_270)  // jesli gracz patrzy w prawo
{
	x_bound=(x/CELL_X_SIZE+1)*CELL_X_SIZE;
	x_delta=CELL_X_SIZE;
	yi=tan_table[view_angle]*(x_bound-x)+y;
	next_x_cell=0;
}
else								// jesli gracz patrzy w lewo
{
	x_bound=(x/CELL_X_SIZE)*CELL_X_SIZE;
	x_delta=-CELL_X_SIZE;
	yi=tan_table[view_angle]*(x_bound-x)+y;
	next_x_cell=-1;
} // Koniec czesci 3

/* Czesc 4 Wyznaczenie dalszych przeciec i odleglosci_x */

syi=yi;
sxi=xi;

while(!xray)
{

	cell_x=(x_bound+next_x_cell)/CELL_X_SIZE;
	cell_y=int(yi)/CELL_Y_SIZE;
	
	
	if(cell_x>=0 && cell_x<WORLD_COLUMNS && cell_y>=0 && cell_y<WORLD_ROWS) // jesli przeciecie wystepuje poza granicami swiata:
		{
		if((bitmap_x=world[WORLD_ROWS-1-cell_y][cell_x])!=0)
			{
			dist_x=float(fabs((yi-y)*(inv_sin_table[view_angle])));
			yi_save=int(yi);
			xb_save=x_bound;
			xray=true; // odleglosc znaleziona =>koniec poszukiwan w przeciec pionowych
			}
		}
	else
		{
		xray=true;
		dist_x=1e+8;
		}
yi+=y_step[view_angle];	
//yi+=step_y;
	x_bound+=x_delta;
}

yi=syi;
xi=sxi;
while(!yray)
{
	
	cell_x=int(xi/CELL_X_SIZE);
	cell_y=(y_bound+next_y_cell)/CELL_Y_SIZE;
	
	if(cell_x>=0 && cell_x<WORLD_COLUMNS && cell_y>=0 && cell_y<WORLD_ROWS) // jesli w swiecie:
		{
		if((bitmap_y=world[WORLD_ROWS-1-cell_y][cell_x])!=0)
			{
			dist_y=float(fabs((xi-x)*(inv_cos_table[view_angle])));
			xi_save=int(xi);
			yb_save=y_bound;
			yray=true; // odleglosc znaleziona =>koniec poszukiwan w przeciec pionowych
			}
		else 
			{
			
			}
		}
	else
		{
		dist_y=1e+8;
		yray=true;
		}

xi+=x_step[view_angle];	
//xi+=step_x;
	y_bound+=y_delta;
}
    

if (dist_x >dist_y) 
{
	dist_array[ray]=float(scale/(1e-10+dist_y)); //wybor mniejszej odleglosci przeciecia
	marked_texture[ray]=bitmap_y;
//	color_array[ray]=xi_save-(xi_save/CELL_X_SIZE)*CELL_X_SIZE;
	if(next_y_cell) bound_array[ray]=63-xi_save&0x3f;
	else bound_array[ray]=xi_save&0x3f;
}
else 
{
	
	dist_array[ray]=float(scale/(1e-10+dist_x));
//	color_array[ray]=yi_save-(yi_save/CELL_Y_SIZE)*CELL_Y_SIZE;
	marked_texture[ray]=bitmap_x;	
	if(next_x_cell) bound_array[ray]=yi_save&0x3f;
	else bound_array[ray]=63-yi_save&0x3f;

	
}
	dist_array[ray]*=correction_table[ray];


view_angle++;
if(view_angle>=ANGLE_360) view_angle=0;			//korekcja kata
}

// Koniec czesci 4


}


/* Ta funkcja sluzy do wyznaczenia odleglosci przeciecia promienia
ze sciana dla kata=0) */
float CRayMap::oblicz_dla_0(long _x, long _y,int num_iter)
{
	int	i=0;
	int komorka_x_gracza=_x/CELL_X_SIZE;
	int komorka_y_gracza=_y/CELL_Y_SIZE;
	int polozenie_w_komorce_x=_x-CELL_X_SIZE*komorka_x_gracza;
	while ((marked_texture[num_iter]=world[WORLD_ROWS-1-komorka_y_gracza][komorka_x_gracza+i])==0) i++;

return float(i*(CELL_X_SIZE+1)-polozenie_w_komorce_x);

}

/* Ta funkcja sluzy do wyznaczenia odleglosci przeciecia promienia
ze sciana dla kata=90) */
float CRayMap::oblicz_dla_90(long _x, long _y,int num_iter)
{
	int i=0;
	int komorka_x_gracza=_x/CELL_X_SIZE;
	int komorka_y_gracza=_y/CELL_Y_SIZE;
	int polozenie_w_komorce_y=_y-CELL_X_SIZE*komorka_y_gracza;
	while ((marked_texture[num_iter]=world[WORLD_ROWS-1-komorka_y_gracza-i][komorka_x_gracza])==0) i++;
return float(i*(CELL_X_SIZE+1)-polozenie_w_komorce_y);
}

/* Ta funkcja sluzy do wyznaczenia odleglosci przeciecia promienia
ze sciana dla kata=180) */
float CRayMap::oblicz_dla_180(long _x, long _y,int num_iter)
{
	int	i=0;
	int komorka_x_gracza=_x/CELL_X_SIZE;
	int komorka_y_gracza=_y/CELL_Y_SIZE;
	int polozenie_w_komorce_x=_x-CELL_X_SIZE*komorka_x_gracza;
	while ((marked_texture[num_iter]=world[WORLD_ROWS-1-komorka_y_gracza][komorka_x_gracza-i])==0) i++;
return float((i-1)*CELL_X_SIZE+polozenie_w_komorce_x);
}

/* Ta funkcja sluzy do wyznaczenia odleglosci przeciecia promienia
ze sciana dla kata=270) */
float CRayMap::oblicz_dla_270(long _x, long _y,int num_iter)
{
	int i=0;
	int komorka_x_gracza=_x/CELL_X_SIZE;
	int komorka_y_gracza=_y/CELL_Y_SIZE;
	int polozenie_w_komorce_y=_y-CELL_X_SIZE*komorka_y_gracza;
	while ((marked_texture[num_iter]=world[WORLD_ROWS-1-komorka_y_gracza+i][komorka_x_gracza])==0) i++;
	return float((i-1)*CELL_X_SIZE+polozenie_w_komorce_y);
}



